package cache

import (
	"context"
	"encoding/json"
	model "git.chillcy.com/golang/colorway/internal/pkg/database/mongo/model/base"
	"go.mongodb.org/mongo-driver/bson"
	primitive "go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo"
	"log"
	"sync"
)

type Cache[T any] interface {
	Filter(data ChangeStreamData) bson.M
	DocumentEqual(a, b T) bool
	GetAllDocuments() []T
}

type Base[T any] struct {
	baseModel model.Model[T]
	documents []T
	rwMutex   *sync.RWMutex
}

func NewBase[T any](baseModel model.Model[T]) *Base[T] {
	return &Base[T]{baseModel: baseModel}
}

func (receiver *Base[T]) StartCache(ctx context.Context, cache Cache[T]) error {
	documents, _, err := receiver.baseModel.Find(ctx, bson.M{})
	if err != nil {
		return err
	}
	receiver.setAllDocuments(documents)
	go func() {
		for {
			changeStream, err := receiver.baseModel.Watch(ctx, mongo.Pipeline{})
			if err != nil {
				log.Println(err)
				continue
			}
			for changeStream.Next(context.TODO()) {
				var changeStreamData ChangeStreamData
				changeBody := changeStream.Current.String()
				err = json.Unmarshal([]byte(changeBody), &changeStreamData)
				if err != nil {
					log.Println(err)
					continue
				}
				filter := cache.Filter(changeStreamData)
				document, err := receiver.baseModel.FindOne(ctx, filter)
				if err != nil {
					log.Println(err)
					continue
				}
				documents := receiver.GetAllDocuments()
				for i := range documents {
					if cache.DocumentEqual(documents[i], document) {
						documents[i] = document
					}
				}
				receiver.setAllDocuments(documents)
			}
			err = changeStream.Close(context.TODO())
			if err != nil {
				log.Println(err)
				continue
			}
		}
	}()
	return nil
}

func (receiver *Base[T]) GetAllDocuments() []T {
	receiver.rwMutex.RLock()
	defer receiver.rwMutex.RUnlock()
	tmp := make([]T, len(receiver.documents))
	copy(tmp, receiver.documents)
	return tmp
}

func (receiver *Base[T]) setAllDocuments(documents []T) {
	receiver.rwMutex.Lock()
	defer receiver.rwMutex.Unlock()
	receiver.documents = documents
}

func (receiver *Base[T]) Filter(data ChangeStreamData) bson.M {
	id, err := primitive.ObjectIDFromHex(data.DocumentKey.Id.Oid)
	if err != nil {
		return nil
	}
	return bson.M{"_id": id}
}

type ChangeStreamData struct {
	DocumentKey struct {
		Id struct {
			Oid string `json:"$oid"`
		} `json:"_id"`
	} `json:"documentKey"`
}
